@@ -1,12 +1,16 @@
 #!/usr/bin/env python3
 # -*- coding: utf-8 -*-
-"""Cons and friends."""
+"""Cons and friends.
+
+Hashable, pickleable, hooks into the built-in reversed(), prints like in Lisps.
+"""
 
 from abc import ABCMeta, abstractmethod
+from collections.abc import Sequence
 from itertools import zip_longest
 
-from unpythonic.fun import composer1
-from unpythonic.it import foldr, foldl
+from unpythonic.fun import composer1i
+from unpythonic.fold import foldr, foldl
 
 # explicit list better for tooling support
 _exports = ["cons", "nil",
@@ -17,7 +21,7 @@ _exports = ["cons", "nil",
             "caaar", "caadr", "cadar", "caddr", "cdaar", "cdadr", "cddar", "cdddr",
             "caaaar", "caaadr", "caadar", "caaddr", "cadaar", "cadadr", "caddar", "cadddr",
             "cdaaar", "cdaadr", "cdadar", "cdaddr", "cddaar", "cddadr", "cdddar", "cddddr",
-            "ll", "ll_from_sequence", "lreverse", "lappend", "lzip"]
+            "ll", "llist", "lreverse", "lappend", "lzip"]
 #from itertools import product, repeat
 #_ads = lambda n: product(*repeat("ad", n))
 #_c2r = ["c{}{}r".format(*x) for x in _ads(2)]
@@ -82,6 +86,15 @@ class LinkedListIterator(ConsIterator):
                         raise TypeError("Not a linked list")
         super().__init__(head, walker)
 
+class ReversedLinkedListIterator(LinkedListIterator):
+    """Iterator for walking a linked list backwards.
+
+    Computes the reversed list at init time, so it can then be walked forward.
+    Cost O(n)."""
+    def __init__(self, head, _fullerror=True):
+        self._data = lreverse(head)
+        super().__init__(self._data, _fullerror)
+
 class LinkedListOrCellIterator(ConsIterator):
     """Like LinkedListIterator, but allow also a single cons cell.
 
@@ -154,8 +167,12 @@ class cons:
             if isinstance(state[k], Nil):
                 state[k] = nil
         self.__dict__ = state
-    def __iter__(self):  # sequence unpacking of single cells and lists
+    def __iter__(self):
+        """Return iterator with default iteration scheme: single cell or list."""
         return LinkedListOrCellIterator(self)
+    def __reversed__(self):
+        """For lists. Caution: O(n), works by building a reversed list."""
+        return ReversedLinkedListIterator(self)
     def __repr__(self):
         try:  # duck test linked list (true list only, no single-cell pair)
             # listcomp, not genexpr, since we want to trigger any exceptions **now**.
@@ -176,73 +193,132 @@ class cons:
                 return self.car == other.car and self.cdr == other.cdr
         return False
     def __hash__(self):
-        return hash((hash(self.car), hash(self.cdr)))
+        try:  # duck test linked list
+            tpl = tuple(self)
+        except TypeError:
+            tpl = (self.car, self.cdr)
+        return hash(tpl)
+
+def _car(x):
+    return _typecheck(x).car
+def _cdr(x):
+    return _typecheck(x).cdr
+def _typecheck(x):
+    if not isinstance(x, cons):
+        raise TypeError("Expected a cons, got {} with value {}".format(type(x), x))
+    return x
+def _build_accessor(name):
+    spec = name[1:-1]
+    f = {'a': _car, 'd': _cdr}
+    return composer1i(f[char] for char in spec)
 
 def car(x):
-    """Return the first half of a cons cell."""
-    if not isinstance(x, cons):
-        raise TypeError("Expected a cons, got {} with value {}".format(type(x), x))
-    return x.car
+    """Return the first half of a cons cell."""  # no autobuild, we want a docstring.
+    return _car(x)
 def cdr(x):
     """Return the second half of a cons cell."""
-    if not isinstance(x, cons):
-        raise TypeError("Expected a cons, got {} with value {}".format(type(x), x))
-    return x.cdr
+    return _cdr(x)
 
-caar = composer1(car, car)
-cadr = composer1(car, cdr)
-cdar = composer1(cdr, car)
-cddr = composer1(cdr, cdr)
+caar = _build_accessor("caar")
+cadr = _build_accessor("cadr")
+cdar = _build_accessor("cdar")
+cddr = _build_accessor("cddr")
 
-caaar = composer1(car, car, car)
-caadr = composer1(car, car, cdr)
-cadar = composer1(car, cdr, car)
-caddr = composer1(car, cdr, cdr)
-cdaar = composer1(cdr, car, car)
-cdadr = composer1(cdr, car, cdr)
-cddar = composer1(cdr, cdr, car)
-cdddr = composer1(cdr, cdr, cdr)
+caaar = _build_accessor("caaar")
+caadr = _build_accessor("caadr")
+cadar = _build_accessor("cadar")  # look, it's Darth Cadar!
+caddr = _build_accessor("caddr")
+cdaar = _build_accessor("cdaar")
+cdadr = _build_accessor("cdadr")
+cddar = _build_accessor("cddar")
+cdddr = _build_accessor("cdddr")
 
-caaaar = composer1(car, car, car, car)
-caaadr = composer1(car, car, car, cdr)
-caadar = composer1(car, car, cdr, car)
-caaddr = composer1(car, car, cdr, cdr)
-cadaar = composer1(car, cdr, car, car)
-cadadr = composer1(car, cdr, car, cdr)
-caddar = composer1(car, cdr, cdr, car)
-cadddr = composer1(car, cdr, cdr, cdr)
-cdaaar = composer1(cdr, car, car, car)
-cdaadr = composer1(cdr, car, car, cdr)
-cdadar = composer1(cdr, car, cdr, car)
-cdaddr = composer1(cdr, car, cdr, cdr)
-cddaar = composer1(cdr, cdr, car, car)
-cddadr = composer1(cdr, cdr, car, cdr)
-cdddar = composer1(cdr, cdr, cdr, car)
-cddddr = composer1(cdr, cdr, cdr, cdr)
+caaaar = _build_accessor("caaaar")
+caaadr = _build_accessor("caaadr")
+caadar = _build_accessor("caadar")
+caaddr = _build_accessor("caaddr")
+cadaar = _build_accessor("cadaar")
+cadadr = _build_accessor("cadadr")
+caddar = _build_accessor("caddar")
+cadddr = _build_accessor("cadddr")
+cdaaar = _build_accessor("cdaaar")
+cdaadr = _build_accessor("cdaadr")
+cdadar = _build_accessor("cdadar")
+cdaddr = _build_accessor("cdaddr")
+cddaar = _build_accessor("cddaar")
+cddadr = _build_accessor("cddadr")
+cdddar = _build_accessor("cdddar")
+cddddr = _build_accessor("cddddr")
 
 def ll(*elts):
-    """Pack elts to a linked list."""
-    return ll_from_sequence(elts)
+    """Make a linked list with the given elements.
 
-def ll_from_sequence(sequence):
-    """Convert sequence to a linked list."""
-    return foldr(cons, nil, sequence)
+    ``ll(...)`` plays the same role as ``[...]`` or ``(...)`` for lists or tuples,
+    respectively, but for linked lists. See also ``llist``.
+
+    **NOTE**: The returned data type is ``cons``, there is no ``ll`` type.
+    A linked list is just one kind of structure that can be built out of cons cells.
+
+    Equivalent to ``(list ...)`` in Lisps. Since in Python the name ``list``
+    refers to the builtin dynamic array type, we use the name ``ll``.
+    """
+    return llist(elts)
+
+def llist(iterable):
+    """Make a linked list from iterable.
+
+    ``llist(...)`` plays the same role as ``list(...)`` or ``tuple(...)`` for
+    lists or tuples, respectively, but for linked lists. See also ``ll``.
+
+    **NOTE**: The returned data type is ``cons``, there is no ``llist`` type.
+    A linked list is just one kind of structure that can be built out of cons cells.
+
+    **Efficiency**:
+
+    Because cons appends to the front, this is efficient for:
+
+      - ``reversed(some_linked_list)``, by just returning the already computed
+        reversed list that is internally stored by the iterator.
+
+      - Sequences, since they can be walked backwards; a linear walk is enough.
+        Here a sequence is defined as tuple, list, range, or any custom class
+        inheriting from ``collections.abc.Sequence``.
+
+    For a general iterable input, this costs a linear walk (forwards), plus an
+    ``lreverse`` once the list has been fully consed.
+    """
+    if isinstance(iterable, ReversedLinkedListIterator):
+        # avoid two extra reverses by reusing the internal data.
+        return iterable._data
+    if isinstance(iterable, (tuple, list, range, Sequence)):
+        return foldr(cons, nil, iterable)  # sequences can be walked backwards
+    # general iterable requires walking forwards, then reversing the result
+    # because cons appends to the front.
+    #
+    # Equivalent to lreverse(lreverse(iterable)) but this is semantically cleaner,
+    # since the original iterable is usually not a linked list. (There's no point
+    # in copying one, since they're immutable.)
+    return lreverse(foldl(cons, nil, iterable))
 
 def lreverse(l):
-    """Reverse a linked list."""
+    """Reverse a linked list, returning the resulting list.
+
+    If you want an iterator instead, use ``reversed(l)``. The computational cost
+    is the same in both cases, O(n).
+    """
     return foldl(cons, nil, l)
 
 def lappend(*ls):
-    """Append linked lists left-to-right."""
+    """Append the given linked lists left-to-right."""
     def lappend_two(l1, l2):
-        return foldr(cons, l2, tuple(l1))  # tuple() because foldr needs a sequence
+        return foldr(cons, l2, l1)  # this internally calls reversed() on l1
     return foldr(lappend_two, nil, ls)
 
 def member(x, l):
-    """Walk linked list and check if item x is in it.
+    """Walk linked list l and check if item x is in it.
 
     Returns:
-        The matching cons cell (tail of list) if x was found; False if not.
+        The matching cons cell (tail of l) if x was found; False if not.
     """
     it = TailIterator(l)
     for t in it:
@@ -255,7 +331,7 @@ def lzip(*ls):
 
     Built-in zip() works too, but produces tuples.
     """
-    return ll(*map(ll, *ls))
+    return llist(map(ll, *ls))
 
 def test():
     # TODO: extend tests
@@ -268,10 +344,13 @@ def test():
 #    print(cons(cons(cons(nil, 3), 2), 1))  # improper list
 
     # type conversion
-    lst = ll(1, 2, 3)
-    assert list(lst) == [1, 2, 3]
-    assert tuple(lst) == (1, 2, 3)
-    assert ll_from_sequence((1, 2, 3)) == lst
+    tpl = (1, 2, 3)
+    lst = [1, 2, 3]
+    llst = ll(1, 2, 3)
+    assert list(llst) == lst
+    assert tuple(llst) == tpl
+    assert llist(tpl) == llst
+    assert llist(lst) == llst
     assert tuple(nil) == ()
 
     # equality
@@ -303,7 +382,7 @@ def test():
         assert False, "binary tree should not be iterable as a linked list"
 
     # should be able to repr() general cons structures
-    repr(t)
+    repr(t)  # should not crash
 
     q = ll(cons(1, 2), cons(3, 4))  # list of pairs, not a tree!
     assert [f(q) for f in [caar, cdar, cadr, cddr]] == [1, 2, cons(3, 4), nil]
@@ -347,7 +426,7 @@ def test():
     l3 = cons(6, cdr(l1))
     assert l3 == ll(6, 2, 1)
 
-    # test hashability
+    # hashability
     s = set()
     s.add(cons(1, 2))
     s.add(ll(1, 2, 3))
@@ -356,6 +435,18 @@ def test():
     assert cons(3, 4) not in s
     assert ll(1, 2) not in s
 
+    # reverse
+    assert llist(reversed(ll(1, 2, 3))) == ll(3, 2, 1)
+    assert foldl(cons, nil, ll(1, 2, 3)) == ll(3, 2, 1)
+
+    # foldr (as implemented in unpythonic) requires support for __reversed__,
+    # since a linked list is an iterable, not a sequence.
+    assert foldr(cons, nil, ll(1, 2, 3)) == ll(1, 2, 3)
+
+    # implementation detail to avoid extra reverses
+    r = reversed(ll(1, 2, 3))   # an iterator that internally builds the reversed list...
+    assert llist(r) is r._data  # ...which llist should just grab
+
     print("All tests PASSED")
 
 if __name__ == '__main__':
